"""
Daemonize a process.

Based on: http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
Taken from: https://github.com/WIPACrepo/iceprod/blob/master/iceprod/server/daemon.py
"""

# fmt:off

import atexit
import os
import signal
import sys
import time


class Daemon(object):
    """
    A generic daemon class.

    Usage:
      d=Daemon(pidfile - filename for pidfile (required)
               runner - function to execute in daemon (required)
               stdin - input filename (default is /dev/null)
               stdout - output filename (default is /dev/null)
               stderr - error filename (default is /dev/null)
               chdir - working directory (default is /)
               umask - umask of new files (default is 0)
              )
      d.start()
      d.stop()
      d.kill()
    """
    def __init__(self, pidfile, runner,
                 stdin='/dev/null',
                 stdout='/dev/null',
                 stderr='/dev/null',
                 chdir='/',
                 umask=0):
        if not isinstance(pidfile, str):
            raise Exception('pidfile is not a string')
        if not callable(runner):
            raise Exception('runner is not callable')
        self.pidfile = pidfile
        self.runner = runner
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.chdir = chdir
        self.umask = umask

    def _daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # decouple from parent environment
        os.chdir(self.chdir)
        os.setsid()
        os.umask(self.umask)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as e:
            sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(self.stdin, 'rb')
        so = open(self.stdout, 'ab+')
        se = open(self.stderr, 'ab+', 0)
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        pgrp = str(os.getpgrp())
        open(self.pidfile,'w+').write("%s %s\n" % (pid,pgrp))

    def _sendsignal(self,pid,sig,waitfordeath=True):
        """Send the specified signal to the process"""
        try:
            os.kill(pid, sig)
            if waitfordeath:
                for _ in range(10):
                    time.sleep(1)
                    os.kill(pid, sig)
                return False
        except OSError as err:
            err = str(err)
            if 'No such process' in err:
                self.delpid()
            else:
                sys.stdout.write("OSError: %s\n" % err)
                sys.exit(1)
        return True

    def _sendsignalgrp(self,pid,sig,waitfordeath=True):
        """Send the specified signal to the process group"""
        try:
            os.killpg(pid, sig)
            if waitfordeath:
                for _ in range(10):
                    time.sleep(1)
                    os.killpg(pid, sig)
                return False
        except OSError as err:
            err = str(err)
            if 'No such process' in err:
                self.delpid()
            else:
                sys.stdout.write("OSError: %s\n" % err)
                sys.exit(1)
        return True

    def delpid(self):
        if os.path.exists(self.pidfile):
            sys.stdout.write("Deleting pidfile\n")
            os.remove(self.pidfile)

    def getpid(self):
        """Get the pid from the pidfile"""
        try:
            pf = open(self.pidfile,'r')
            pid,pgrp = [int(x.strip()) for x in pf.read().split()]
            pf.close()
        except IOError:
            pid = None
            pgrp = None
        return (pid,pgrp)

    def start(self):
        """Start the daemon"""
        pid,pgrp = self.getpid()
        if pid:
            message = "pidfile %s already exist. Daemon already running?"
            raise Exception(message % self.pidfile)

        # Start the daemon
        self._daemonize()
        self.runner()

    def stop(self):
        """Stop the daemon"""
        pid,pgrp = self.getpid()
        if not pid:
            message = "pidfile %s does not exist. Daemon not running?\n"
            sys.stderr.write(message % self.pidfile)
            return
        if not self._sendsignal(pid,signal.SIGINT):
            sys.stdout.write('SIGINT failed, try SIGQUIT\n')
            if not self._sendsignalgrp(pgrp,signal.SIGQUIT):
                sys.stdout.write('SIGQUIT failed, try SIGKILL\n')
                self._sendsignalgrp(pgrp,signal.SIGKILL)

    def kill(self):
        """Kill the daemon"""
        pid,pgrp = self.getpid()
        if not pid:
            message = "pidfile %s does not exist. Daemon not running?\n"
            sys.stderr.write(message % self.pidfile)
            return
        if not self._sendsignal(pid,signal.SIGQUIT):
            self._sendsignalgrp(pgrp,signal.SIGTERM)

    def hardkill(self):
        """Kill the daemon"""
        pid,pgrp = self.getpid()
        if not pid:
            message = "pidfile %s does not exist. Daemon not running?\n"
            sys.stderr.write(message % self.pidfile)
            return
        self._sendsignalgrp(pgrp,signal.SIGKILL)

    def restart(self):
        """Restart the daemon"""
        self.stop()
        self.start()



"""Get a `requests`_ Session that fully retries errors.

.. _requests: http://docs.python-requests.org
"""

# fmt:off
# pylint: skip-file

from typing import Iterable

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry  # type: ignore[import]
from requests_futures.sessions import FuturesSession  # type: ignore[import]


def AsyncSession(
    retries: int = 10,
    backoff_factor: float = 0.3,
    allowed_methods: Iterable[str] = ('HEAD', 'TRACE', 'GET', 'POST', 'PUT', 'OPTIONS', 'DELETE'),
    status_forcelist: Iterable[int] = (408, 429, 500, 502, 503, 504),
) -> FuturesSession:
    """Return a Session object with full retry capabilities.

    Args:
        retries (int): number of retries
        backoff_factor (float): speed factor for retries (in seconds)
        allowed_methods (iterable): http methods to retry on
        status_forcelist (iterable): http status codes to retry on

    Returns:
        :py:class:`requests.Session`: session object
    """
    session = FuturesSession()
    retry = Retry(
        total=retries,
        connect=retries,
        read=retries,
        redirect=retries,
        # status=retries,
        allowed_methods=allowed_methods,
        status_forcelist=status_forcelist,
        backoff_factor=backoff_factor,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def Session(
    retries: int = 10,
    backoff_factor: float = 0.3,
    allowed_methods: Iterable[str] = ('HEAD', 'TRACE', 'GET', 'POST', 'PUT', 'OPTIONS', 'DELETE'),
    status_forcelist: Iterable[int] = (408, 429, 500, 502, 503, 504),
) -> requests.Session:
    """Return a Session object with full retry capabilities.

    Args:
        retries (int): number of retries
        backoff_factor (float): speed factor for retries (in seconds)
        allowed_methods (iterable): http methods to retry on
        status_forcelist (iterable): http status codes to retry on

    Returns:
        :py:class:`requests.Session`: session object
    """
    session = requests.Session()
    retry = Retry(
        total=retries,
        connect=retries,
        read=retries,
        redirect=retries,
        # status=retries,
        allowed_methods=allowed_methods,
        status_forcelist=status_forcelist,
        backoff_factor=backoff_factor,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session



# fmt:off
# pylint: skip-file

import json
import logging
import time

import jwt
import requests


class _AuthValidate:
    def __init__(self, audience=None, issuers=None, algorithms=None):
        self.audience = audience
        self.issuers = issuers
        self.algorithms = algorithms if algorithms else ['RS256','RS512']

    def _validate(self, token, key, **kwargs):
        options = {}

        # required claims
        claims = ['exp', 'iat', 'iss']
        claims.extend(kwargs.pop('required', []))
        options['require'] = claims

        # configure the audience to validate
        audience = kwargs.pop('audience', None)
        if not audience:
            audience = self.audience
        if not audience:
            options['verify_aud'] = False
        else:
            if isinstance(audience, str):
                audience = [audience]
            kwargs['audience'] = audience

        # configure issuers to validate
        issuers = kwargs.pop('issuers', None)
        if not issuers:
            iss = kwargs.pop('issuer', None)
            if iss:
                issuers = [iss]
            else:
                issuers = self.issuers

        token = jwt.decode(token, key, algorithms=self.algorithms, options=options, **kwargs)
        if issuers and token['iss'] not in issuers:
            raise jwt.exceptions.InvalidIssuerError()

        return token


class Auth(_AuthValidate):
    """
    Handle authentication of JWT tokens.
    """

    def __init__(self, secret, pub_secret=None, issuer='IceProd', algorithm='HS512', expiration=31622400, expiration_temp=86400, **kwargs):
        if 'algorithms' not in kwargs:
            kwargs['algorithms'] = [algorithm]
        super().__init__(**kwargs)
        self.secret = secret
        self.pub_secret = pub_secret if pub_secret else secret
        self.issuer = issuer
        self.algorithm = algorithm
        self.max_exp = expiration
        self.max_exp_temp = expiration_temp

    def create_token(self, subject, expiration=None, type='temp', payload=None, headers=None):
        """
        Create a token.

        Args:
            subject (str): the user or other owner
            expiration (int): duration in seconds for which the token is valid
            type (str): type of token
            payload (dict): any other fields that should be included

        Returns:
            str: representation of JWT token
        """
        exp = self.max_exp_temp if type == 'temp' else self.max_exp
        if expiration and exp > expiration:
            exp = expiration
        now = time.time()
        if not payload:
            payload = {}
        payload.update({
            'iss': self.issuer,
            'sub': subject,
            'exp': now+exp,
            'nbf': now,
            'iat': now,
            'type': type,
        })

        token = jwt.encode(payload, self.secret, algorithm=self.algorithm, headers=headers)
        return token

    def validate(self, token, **kwargs):
        """
        Validate a token.

        Args:
            token (str): a JWT token
            **kwargs: additional args passed directly to jwt.decode

        Returns:
            dict: data inside token

        Raises:
            Exception on failure to validate.
        """
        return self._validate(token, self.pub_secret, issuer=self.issuer, required=['type'], **kwargs)


class OpenIDAuth(_AuthValidate):
    """Handle validation of JWT tokens using OpenID .well-known auto-discovery."""
    def __init__(self, url, **kwargs):
        super().__init__(**kwargs)
        self.url = url if url.endswith('/') else url+'/'
        self.public_keys = {}
        self.provider_info = {}
        self.token_url = None

        self._refresh_keys()

    def _refresh_keys(self):
        try:
            if not self.provider_info:
                # discovery
                r = requests.get(self.url+'.well-known/openid-configuration')
                r.raise_for_status()
                self.provider_info = r.json()

                # get token url
                self.token_url = self.provider_info['token_endpoint']

            # get keys
            r = requests.get(self.provider_info['jwks_uri'])
            r.raise_for_status()
            for jwk in r.json()['keys']:
                logging.debug(f'jwk: {jwk}')
                kid = jwk['kid']
                logging.info(f'loaded JWT key {kid}')
                self.public_keys[kid] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))
        except Exception:
            logging.warning('failed to refresh OpenID keys', exc_info=True)

    def validate(self, token, **kwargs):
        """
        Validate a token.

        Args:
            token (str): a JWT token
            audience (str): audience, or None to disable audience verification

        Returns:
            dict: data inside token

        Raises:
            Exception on failure to validate.
        """
        header = jwt.get_unverified_header(token)
        if header['kid'] not in self.public_keys:
            self._refresh_keys()
        if header['kid'] in self.public_keys:
            key = self.public_keys[header['kid']]
            return self._validate(token, key, **kwargs)
        else:
            raise Exception(f'JWT key {header["kid"]} not found')



"""RestHandler and related classes."""

# fmt:off
# pylint: skip-file

import base64
import hashlib
import hmac
import logging
import secrets
import time
import urllib.parse
from collections import defaultdict
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    List,
    MutableMapping,
    Optional,
    Type,
    TypeVar,
    Union,
)

import rest_tools
import tornado.escape
import tornado.gen
import tornado.httpclient
import tornado.httputil
import tornado.web
from cachetools import TTLCache
from tornado.auth import OAuth2Mixin

from .. import telemetry as wtt
from ..utils.auth import Auth, OpenIDAuth
from . import arghandler
from .decorators import catch_error
from .stats import RouteStats

T = TypeVar("T")


logger = logging.getLogger('rest')


def RestHandlerSetup(config={}):
    """Default RestHandler setup. Returns the default arguments for passing to
    the route setup.

    Args:
        config (dict): config dict

    Returns:
        dict: handler config
    """
    debug = True if 'debug' in config and config['debug'] else False
    auth = None
    auth_url = ''
    module_auth_key = ''
    if 'auth' in config:
        kwargs = {}
        if 'audience' in config['auth']:
            kwargs['audience'] = config['auth']['audience']
        if 'issuers' in config['auth']:
            kwargs['issuers'] = config['auth']['issuers']
        if 'secret' in config['auth']:
            kwargs['secret'] = config['auth']['secret']
            if 'issuer' in config['auth']:
                kwargs['issuer'] = config['auth']['issuer']
            if 'algorithm' in config['auth']:
                kwargs['algorithm'] = config['auth']['algorithm']
            if 'expiration' in config['auth']:
                kwargs['expiration'] = config['auth']['expiration']
            if 'expiration_temp' in config['auth']:
                kwargs['expiration_temp'] = config['auth']['expiration_temp']
            auth = Auth(**kwargs)
        elif 'openid_url' in config['auth']:
            if 'algorithms' in config['auth']:
                kwargs['algorithms'] = config['auth']['algorithms']
            auth = OpenIDAuth(config['auth']['openid_url'], **kwargs)
            if auth.token_url:
                auth_url = auth.token_url
        if 'url' in config['auth']:
            auth_url = config['auth']['url']
    if 'rest_api' in config and 'auth_key' in config['rest_api']:
        module_auth_key = config['rest_api']['auth_key']

    if 'route_stats' in config:
        route_stats = defaultdict(partial(RouteStats, **config['route_stats']))
    else:
        route_stats = defaultdict(RouteStats)

    return {
        'debug': debug,
        'auth': auth,
        'auth_url': auth_url,
        'module_auth_key': module_auth_key,
        'server_header': config.get('server_header', 'REST'),
        'route_stats': route_stats
    }


class RestHandler(tornado.web.RequestHandler):
    """Default REST handler."""
    def __init__(self, *args, **kwargs) -> None:
        self.server_header = ''
        try:
            super().__init__(*args, **kwargs)
        except Exception:
            logging.error('error', exc_info=True)

    def initialize(self, debug=False, auth=None, auth_url=None, module_auth_key='', server_header='', route_stats=None, **kwargs):
        super().initialize(**kwargs)
        self.debug = debug
        self.auth = auth
        self.auth_url = auth_url
        self.auth_data = {}
        self.auth_key = None
        self.module_auth_key = module_auth_key
        self.server_header = server_header
        self.route_stats = route_stats

    @wtt.spanned(
        span_namer=wtt.SpanNamer(use_this_arg='self.request.method'),
        kind=wtt.SpanKind.SERVER,
        these=["self.request.method", "self.request.path"],
        carrier="self.request.headers",
    )
    async def _execute(self, *args: Any, **kwargs: Any) -> None:
        """Call implemented methods.

        NOTE: This is the closest common call-stack ancestor of
            - `prepare()`,
            - "method handlers" (`get()`, `post()`, ...),
            - `on_finish()`,
            - etc.
        """
        return await super()._execute(*args, **kwargs)

    def set_default_headers(self):
        self._headers['Server'] = self.server_header

    def get_template_namespace(self):
        namespace = super().get_template_namespace()
        namespace['version'] = rest_tools.__version__
        return namespace

    def get_current_user(self):
        """Get the current user, and set auth-related attributes."""
        try:
            type, token = self.request.headers['Authorization'].split(' ', 1)
            if type.lower() != 'bearer':
                raise Exception('bad header type')
            logger.debug('token: %r', token)
            data = self.auth.validate(token)
            self.auth_data = data
            self.auth_key = token
            return data['sub']
        # Auth Failed
        except Exception:
            if self.debug and 'Authorization' in self.request.headers:
                logger.info('Authorization: %r', self.request.headers['Authorization'])
            logger.info('failed auth', exc_info=True)

        return None

    @wtt.evented()
    def prepare(self):
        """Prepare before http-method request handlers."""
        logger.debug(f"{self.request.method} [{self.__class__.__name__}]")

        if self.route_stats is not None:
            stat = self.route_stats[self.request.path]
            if stat.is_overloaded():
                backoff = stat.get_backoff_time()
                logger.warn('Server is overloaded, backoff %r', backoff)
                self.set_header('Retry-After', backoff)
                raise tornado.web.HTTPError(503, reason="server overloaded")
            self.start_time = time.time()

    @wtt.evented()
    def on_finish(self):
        """Cleanup after http-method request handlers."""
        if self.route_stats is not None and self.get_status() < 500:
            stat = self.route_stats[self.request.path]
            stat.append(time.time() - self.start_time)

    @wtt.evented(all_args=True)
    def write_error(self, status_code=500, **kwargs):
        """Write out custom error json."""
        data = {
            'code': status_code,
            'error': self._reason,
        }
        self.write(data)
        self.finish()

    def get_json_body_argument(
        self,
        name: str,
        default: Any = arghandler.NO_DEFAULT,
        type: Union[None, Type[T], Callable[[Any],T]] = None,
        choices: Optional[List[Any]] = None,
        forbiddens: Optional[List[Any]] = None,
        strict_type: bool = False,
    ) -> T:
        """Get argument from the JSON-decoded request-body.

        If no `default` is provided, and the argument is not present, raise `400`.

        Arguments:
            name -- the argument's name

        Keyword Arguments:
            default -- a default value to use if the argument is not present
            type -- typecast (or call a one-argument callable) with the argument's value (raise `400` for ValueError and/or TypeError)
            choices -- a list of valid argument values; regex is allowed for strings (raise `400`, if arg's value is not in list)
            forbiddens -- a list of disallowed argument values; regex is allowed for strings (raise `400`, if arg's value is in list)
            strict_type -- if True and `type` is passed, the arg's value is type-checked instead of type-casted

        Returns:
            the argument's value, possibly stripped/typecasted
        """
        return arghandler.ArgumentHandler.get_json_body_argument(
            self.request.body, name, default, type, choices, forbiddens, strict_type
        )

    def get_argument(
        self,
        name: str,
        default: Any = arghandler.NO_DEFAULT,
        strip: bool = True,
        type: Union[None, Type[T], Callable[[Any],T]] = None,
        choices: Optional[List[Any]] = None,
        forbiddens: Optional[List[Any]] = None,
        strict_type: bool = False,
    ) -> T:
        """Get argument from query base-arguments / JSON-decoded request-body.

        If no `default` is provided, and the argument is not present, raise `400`.

        Arguments:
            name -- the argument's name

        Keyword Arguments:
            default -- a default value to use if the argument is not present
            strip {`bool`} -- whether to `str.strip()` the arg's value (default: {`True`})
            type -- typecast (or call a one-argument callable) with the argument's value (raise `400` for ValueError and/or TypeError)
            choices -- a list of valid argument values; regex is allowed for strings (raise `400`, if arg's value is not in list)
            forbiddens -- a list of disallowed argument values; regex is allowed for strings(raise `400`, if arg's value is in list)
            strict_type -- if True and `type` is passed, the arg's value is type-checked instead of type-casted

        Returns:
            the argument's value, possibly stripped/typecasted
        """
        return arghandler.ArgumentHandler.get_argument(
            self.request.body,
            super().get_argument,
            name,
            default,
            strip,
            type,
            choices,
            forbiddens,
            strict_type,
        )


class KeycloakUsernameMixin:
    """Get the username correctly from Keycloak tokens.

    Note: will not work on service account tokens.
    """
    def get_current_user(self):
        if not super().get_current_user():
            return None
        username = self.auth_data.get('preferred_username', None)
        if not username:
            username = self.auth_data.get('upn', None)
            if not username:
                logger.info('could not find auth username')
        return username


class OpenIDCookieHandlerMixin:
    """Store/load current user's `OpenIDLoginHandler` tokens in cookies."""
    def get_current_user(self):
        """Get the current user, and set auth-related attributes."""
        try:
            access_token = self.get_secure_cookie('access_token')
            data = self.auth.validate(access_token)
            self.auth_data = data
            self.auth_key = access_token
            refresh_token = self.get_secure_cookie('refresh_token')
            self.auth_refresh_token = refresh_token
            return data['sub']
        # Auth Failed
        except Exception:
            logger.debug('failed auth', exc_info=True)

        return None

    def store_tokens(
        self,
        access_token,
        access_token_exp,
        refresh_token=None,
        refresh_token_exp=None,
        user_info=None,
        user_info_exp=None,
    ):
        """
        Store jwt tokens and user info from OpenID-compliant auth source.

        Args:
            access_token (str): jwt access token
            access_token_exp (int): access token expiration in seconds
            refresh_token (str): jwt refresh token
            refresh_token_exp (int): refresh token expiration in seconds
            user_info (dict): user info (from id token or user info lookup)
            user_info_exp (int): user info expiration in seconds
        """
        self.set_secure_cookie('access_token', access_token,
                               expires_days=float(access_token_exp)/3600/24)
        if refresh_token and refresh_token_exp:
            self.set_secure_cookie('refresh_token', refresh_token,
                                   expires_days=float(refresh_token_exp)/3600/24)
        if user_info and user_info_exp:
            self.set_secure_cookie('user_info', tornado.escape.json_encode(user_info),
                                   expires_days=float(user_info_exp)/3600/24)

    def clear_tokens(self):
        """
        Clear token data, usually on logout.
        """
        self.clear_cookie('access_token')
        self.clear_cookie('refresh_token')
        self.clear_cookie('user_info')


class OpenIDLoginHandler(OpenIDCookieHandlerMixin, OAuth2Mixin, RestHandler):
    """Handle OpenID Connect logins.

    Should be combined with an appropriate mixin to store the token(s).
    `OpenIDCookieHandlerMixin` is used by default, but can be overridden.

    Requires the `login_url` application setting to be a full url.
    """
    _pkcs_challenges: MutableMapping[str, str] = TTLCache(maxsize=10000, ttl=3600)

    def initialize(self, oauth_client_id, oauth_client_secret, oauth_client_scope=None, **kwargs):
        super().initialize(**kwargs)
        if not isinstance(self.auth, OpenIDAuth):
            raise RuntimeError('OpenID Connect auth not set up')
        self._OAUTH_AUTHORIZE_URL = self.auth.provider_info['authorization_endpoint']
        self._OAUTH_ACCESS_TOKEN_URL = self.auth.provider_info['token_endpoint']
        self._OAUTH_LOGOUT_URL = self.auth.provider_info['end_session_endpoint']
        self._OAUTH_USERINFO_URL = self.auth.provider_info['userinfo_endpoint']
        self.oauth_client_id = oauth_client_id
        self.oauth_client_secret = oauth_client_secret
        if oauth_client_scope:
            self.oauth_client_scope = oauth_client_scope.split()
        else:
            self.oauth_client_scope = ['profile', 'groups']
            if oauth_client_secret:
                self.oauth_client_scope.append('offline_access')

    @classmethod
    def create_pkce_challenge(cls) -> str:
        code_verifier = secrets.token_urlsafe(64)
        code_challenge = base64.urlsafe_b64encode(hashlib.sha256(code_verifier.encode('utf-8')).digest()).decode('utf-8').split('=')[0]
        cls._pkcs_challenges[code_challenge] = code_verifier
        return code_challenge

    @classmethod
    def get_pkce_verifier(cls, challenge: str) -> str:
        if challenge in cls._pkcs_challenges:
            return cls._pkcs_challenges[challenge]
        raise KeyError('invalid pkce challenge')

    async def get_authenticated_user(
        self, redirect_uri: str, code: str, state: Dict[str, Any]
    ) -> Dict[str, Any]:
        http = self.get_auth_http_client()
        body = {
            'redirect_uri': redirect_uri,
            'code': code,
            'client_id': self.oauth_client_id,
            'grant_type': 'authorization_code',
        }
        if self.oauth_client_secret:
            body['client_secret'] = self.oauth_client_secret
        else:
            # use PKCE
            code_challenge = state.get('code_challenge', None)
            if not code_challenge:
                raise tornado.web.HTTPError(400, reason='missing PKCE code challenge')
            body['code_verifier'] = self.get_pkce_verifier(code_challenge)

        response = await http.fetch(
            self._OAUTH_ACCESS_TOKEN_URL,
            method='POST',
            headers={'Content-Type': 'application/x-www-form-urlencoded'},
            body=urllib.parse.urlencode(body),
        )
        ret = tornado.escape.json_decode(response.body)
        if not ret.get('id_token', ''):
            response = await http.fetch(
                self._OAUTH_USERINFO_URL,
                method='GET',
                headers={'Authorization': f'Bearer {ret["access_token"]}'},
            )
            ret['id_token'] = tornado.escape.json_decode(response.body)

        try:
            self.auth.validate(ret['access_token'])
        except Exception:
            if self.debug:
                logger.debug(f'bad token: {ret}')
            raise
        return ret

    def _decode_state(self, state: Union[bytes, str]) -> str:
        data = tornado.escape.json_decode(base64.b64decode(state))
        _, token, _ = self._decode_xsrf_token(data.pop('xsrf'))
        _, expected_token, _ = self._get_raw_xsrf_token()
        if not token:
            raise tornado.web.HTTPError(403, reason="'state' argument has invalid format")
        if not hmac.compare_digest(tornado.escape.utf8(token), tornado.escape.utf8(expected_token)):
            raise tornado.web.HTTPError(403, reason="XSRF cookie does not match state argument")
        return data

    def _encode_state(self, data: Dict[str, Any]) -> bytes:
        data2 = data.copy()  # make a copy to not add xsrf to source dict
        data2['xsrf'] = self.xsrf_token.decode('utf-8')
        return base64.b64encode(tornado.escape.json_encode(data2).encode('utf-8'))

    @catch_error
    async def get(self):
        if self.get_argument('error', False):
            err = self.get_argument('error_description', None)
            if not err:
                err = 'unknown oauth2 error'
            raise tornado.web.HTTPError(400, reason=err)
        elif self.get_argument('code', False):
            data = self._decode_state(self.get_argument('state'))
            user = await self.get_authenticated_user(
                redirect_uri=self.get_login_url(),
                code=self.get_argument('code'),
                state=data,
            )

            # set expire times (can be 0 in user data, which is an invalid cookie)
            access_expire = user.get('expires_in', 0)
            if not access_expire:
                access_expire = 1800
            refresh_expire = user.get('refresh_expires_in', 0)
            if not refresh_expire:
                refresh_expire = 86400

            # Save the user data
            self.store_tokens(
                access_token=user['access_token'],
                access_token_exp=access_expire,
                refresh_token=user.get('refresh_token'),
                refresh_token_exp=refresh_expire,
                user_info=user.get('id_token'),
                user_info_exp=refresh_expire if user.get('refresh_token') else access_expire,
            )

            if data.get('redirect', None):
                url = data['redirect']
                if 'state' in data:
                    url = tornado.httputil.url_concact(url, {'state': data['state']})
                self.redirect(url)
            elif self.settings.get('debug', False):
                self.write(user)
            else:
                raise tornado.web.HTTPError(400, reason='missing redirect')
        else:
            state = {}
            redirect = self.get_argument('next', None)
            if not redirect:
                redirect = self.get_argument('redirect', None)
            if redirect:
                state['redirect'] = redirect
            elif not self.settings.get('debug', False):
                raise tornado.web.HTTPError(400, 'missing redirect')
            if self.get_argument('state', False):
                state['state'] = self.get_argument('state')
            extra_params = {}
            if not self.oauth_client_secret:
                # use PKCE  https://www.rfc-editor.org/rfc/rfc7636
                extra_params['code_challenge'] = self.create_pkce_challenge()
                extra_params['code_challenge_method'] = 'S256'
                state['code_challenge'] = extra_params['code_challenge']
            extra_params["state"] = self._encode_state(state)

            self.authorize_redirect(
                redirect_uri=self.get_login_url(),
                client_id=self.oauth_client_id,
                scope=self.oauth_client_scope,
                extra_params=extra_params,
                response_type='code')



"""
Helpers for setting up `Tornado <http://tornado.readthedocs.io>`_ servers.
"""

# fmt:off
# pylint: skip-file

import binascii
import logging
import socket

import tornado.web
from tornado.platform.asyncio import AsyncIOMainLoop

AsyncIOMainLoop().install()


def tornado_logger(handler):
    """Log tornado messages to our logger"""
    if handler.get_status() < 400:
        log_method = logging.debug
    elif handler.get_status() < 500:
        log_method = logging.warning
    else:
        log_method = logging.error
    request_time = 1000.0 * handler.request.request_time()
    log_method("%d %s %.2fms", handler.get_status(), handler._request_summary(), request_time)


class RestServer:
    def __init__(self, log_function=None, cookie_secret=None, max_body_size=None, **kwargs):
        self.routes = []
        self.http_server = None
        self.max_body_size = None
        self.app_args = dict(kwargs)

        if log_function:
            self.app_args['log_function'] = log_function
        else:
            self.app_args['log_function'] = tornado_logger

        if cookie_secret:
            self.app_args['xsrf_cookies'] = True
            self.app_args['cookie_secret'] = binascii.unhexlify(cookie_secret)

    def add_route(self, *args):
        self.routes.append(tuple(args))

    def startup(self, address='localhost', port=8080):
        """
        Start up a Tornado server.

        Note that after calling this method you still need to call
        :code:`IOLoop.current().start()` to start the server.

        Args:
            app (:py:class:`tornado.web.Application`): Tornado application
            address (str): bind address
            port (int): bind port
        """
        logging.warning('tornado bound to %s:%d', address, port)

        app = tornado.web.Application(self.routes, **self.app_args)

        if self.http_server:
            self.http_server.stop()
        self.http_server = tornado.httpserver.HTTPServer(app, xheaders=True, max_body_size=self.max_body_size)
        self.http_server.bind(port, address=address, family=socket.AF_INET)
        self.http_server.start()

    async def stop(self):
        self.http_server.stop()
        await self.http_server.close_all_connections()



# fmt:off
import io
from itertools import zip_longest
import logging
from pathlib import Path
import time
from typing import Dict, List, Optional

import qrcode  # type: ignore[import]
import requests

from ..utils.auth import OpenIDAuth
from .openid_client import OpenIDRestClient


def _print_qrcode(req: Dict[str, str]) -> None:
    if 'verification_uri_complete' not in req:
        req['verification_uri_complete'] = req['verification_uri']+'?user_code='+req['user_code']

    qr = qrcode.QRCode(border=2)
    qr.add_data(req['verification_uri_complete'])
    f = io.StringIO()
    qr.print_ascii(out=f)
    f.seek(0)
    code = f.read().rstrip().split('\n')
    code_width = max(len(c) for c in code)
    text_width = max(len(req['verification_uri']), 30)
    box_width = 2 + text_width + code_width

    qrtext = f'''
To complete authorization,
scan the QR code or, using
a browser on another device,
visit:
{req["verification_uri"]}

And enter the code:
{req["user_code"]}
'''.split('\n')

    if box_width < 78:
        print('+', '-' * box_width, '+', sep='')
        for text,qrdata in zip_longest(qrtext, code, fillvalue=''):
            print('|  ', f'{text:<{text_width}}', f'{qrdata:{code_width}}', '|', sep='')
        print('+', '-' * box_width, '+', sep='')
    else:
        box_width = max(text_width+4, code_width)
        print('+', '-' * box_width, '+', sep='')
        for text in qrtext:
            print('|  ', f'{text:<{box_width-4}}', '  |', sep='')
        for qrdata in code:
            print('|', f'{qrdata:<{box_width}}', '|', sep='')
        print('+', '-' * box_width, '+', sep='')


def _perform_device_grant(
    logger: logging.Logger,
    device_url: str,
    token_url: str,
    client_id: str,
    client_secret: Optional[str] = None,
    scopes: Optional[List[str]] = None,
) -> str:
    args = {
        'client_id': client_id,
        'scope': 'offline_access ' + (' '.join(scopes) if scopes else ''),
    }
    if client_secret:
        args['client_secret'] = client_secret

    try:
        r = requests.post(device_url, data=args)
        r.raise_for_status()
        req = r.json()
    except requests.exceptions.HTTPError as exc:
        logger.debug('%r', exc.response.text)
        try:
            req = exc.response.json()
        except Exception:
            req = {}
        error = req.get('error', '')
        raise RuntimeError(f'Device authorization failed: {error}') from exc
    except Exception as exc:
        raise RuntimeError('Device authorization failed') from exc

    logger.debug('Device auth in progress')

    _print_qrcode(req)

    args = {
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
        'device_code': req['device_code'],
        'client_id': client_id,
    }
    if client_secret:
        args['client_secret'] = client_secret

    sleep_time = int(req.get('interval', 5))
    while True:
        time.sleep(sleep_time)
        try:
            r = requests.post(token_url, data=args)
            r.raise_for_status()
            req = r.json()
        except requests.exceptions.HTTPError as exc:
            logger.debug('%r', exc.response.text)
            try:
                req = exc.response.json()
            except Exception:
                req = {}
            error = req.get('error', '')
            if error == 'authorization_pending':
                continue
            elif error == 'slow_down':
                sleep_time += 5
                continue
            raise RuntimeError(f'Device authorization failed: {error}') from exc
        except Exception as exc:
            raise RuntimeError('Device authorization failed') from exc
        break

    return req['refresh_token']


def DeviceGrantAuth(
    address: str, token_url: str, client_id: str,
    client_secret: Optional[str] = None,
    scopes: Optional[List[str]] = None,
) -> OpenIDRestClient:
    """A REST client that can handle OpenID and the OAuth2 Device Client flow.

    Args:
        address (str): base address of REST API
        token_url (str): base address of token service
        client_id (str): client id
        client_secret (str): client secret (optional - required for confidential clients)
        scopes (list): token scope list (optional)
    """
    logger = logging.getLogger('DeviceGrantAuth')

    auth = OpenIDAuth(token_url)
    if 'device_authorization_endpoint' not in auth.provider_info:
        raise RuntimeError('Device grant not supported by server')
    endpoint = auth.provider_info['device_authorization_endpoint']

    refresh_token = _perform_device_grant(logger, endpoint, auth.token_url, client_id, client_secret, scopes)

    return OpenIDRestClient(address=address, token_url=token_url, client_id=client_id,
                            client_secret=client_secret, refresh_token=refresh_token)


def _load_token_from_file(filepath: Path) -> Optional[str]:
    if filepath.exists():
        return filepath.read_text()
    return None


def _save_token_to_file(filepath: Path, token: str) -> None:
    filepath.write_text(token)


def SavedDeviceGrantAuth(
    address: str, token_url: str,
    filename: str,
    client_id: str,
    client_secret: Optional[str] = None,
    scopes: Optional[List[str]] = None,
) -> OpenIDRestClient:
    """
    A REST client that can handle OpenID and the OAuth2 Device Client flow,
    and saves a refresh token to a file for quick access.

    Args:
        address (str): base address of REST API
        token_url (str): base address of token service
        filename (str): name of file to save/load refresh token
        client_id (str): client id
        client_secret (str): client secret (optional - required for confidential clients)
        scopes (list): token scope list (optional)
    """
    logger = logging.getLogger('SavedDeviceGrantAuth')

    filepath = Path(filename)
    refresh_token = _load_token_from_file(filepath)

    if not refresh_token:
        auth = OpenIDAuth(token_url)
        if not auth.provider_info:
            raise RuntimeError('Token service does not support .well-known discovery')
        if 'device_authorization_endpoint' not in auth.provider_info:
            raise RuntimeError('Device grant not supported by server')
        endpoint = auth.provider_info['device_authorization_endpoint']
        refresh_token = _perform_device_grant(logger, endpoint, auth.token_url, client_id, client_secret, scopes)

    def update_func(access, refresh):
        _save_token_to_file(filepath, refresh)

    return OpenIDRestClient(address=address, token_url=token_url, client_id=client_id,
                            client_secret=client_secret, refresh_token=refresh_token,
                            update_func=update_func)
